from ride import lifecycle  # noqa: F401


class TestClassificationLifecycle:
    def test_validate_attributes(self):
        """Test validation of attributes"""
        assert False

    def test_metrics_step(self):
        assert False

    def test_metrics_epoch(self):
        assert False

    def test_training_step(self):
        assert False

    def test_training_epoch_end(self):
        assert False

    def test_validation_step(self):
        assert False

    def test_validation_epoch_end(self):
        assert False

    def test_test_step(self):
        assert False

    def test_test_epoch_end(self):
        assert False



import io
import os
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Union

import pytorch_lightning as pl
from matplotlib.figure import Figure
from PIL import Image
from pytorch_lightning.loggers import (
    LightningLoggerBase,
    LoggerCollection,
    NeptuneLogger,
    TensorBoardLogger,
    WandbLogger,
)
from pytorch_lightning.utilities import rank_zero_only

from ride.utils.env import RUN_LOGS_PATH
from ride.utils.io import dump_yaml
from ride.utils.logging import getLogger, process_rank

logger = getLogger(__name__)
ExperimentLogger = Union[TensorBoardLogger, LoggerCollection, WandbLogger]
ExperimentLoggerCreator = Callable[[str], ExperimentLogger]


def singleton_experiment_logger() -> ExperimentLoggerCreator:
    _loggers = {}

    def experiment_logger(
        name: str = None,
        logging_backend: str = "tensorboard",
        Module=None,
        save_dir=RUN_LOGS_PATH,
    ) -> ExperimentLogger:
        nonlocal _loggers
        if logging_backend not in _loggers:
            if process_rank != 0:  # pragma: no cover
                _loggers[logging_backend] = pl.loggers.base.DummyLogger()
                _loggers[logging_backend].log_dir = None
                return _loggers[logging_backend]

            logging_backend = logging_backend.lower()
            if logging_backend == "tensorboard":
                _loggers[logging_backend] = TensorBoardLogger(
                    save_dir=save_dir, name=name
                )
            elif logging_backend == "wandb":
                _loggers[logging_backend] = WandbLogger(
                    save_dir=save_dir,
                    name=name,
                    project=(Module.__name__ if Module else None),
                )
                _loggers[logging_backend].log_dir = getattr(
                    _loggers[logging_backend].experiment._settings, "_sync_dir", None
                )
            else:
                logger.warn("No valid logger selected.")

        return _loggers[logging_backend]

    return experiment_logger


experiment_logger = singleton_experiment_logger()


def fig2img(fig):
    """Convert a Matplotlib figure to a PIL Image and return it"""
    buf = io.BytesIO()
    fig.savefig(buf)
    buf.seek(0)
    img = Image.open(buf)
    return img


def add_experiment_logger(
    prev_logger: LightningLoggerBase, new_logger: LightningLoggerBase
) -> LoggerCollection:
    # If no logger existed previously don't do anything
    if not prev_logger:
        return None

    if isinstance(prev_logger, LoggerCollection):
        return LoggerCollection([*prev_logger._logger_iterable, new_logger])
    else:
        return LoggerCollection([prev_logger, new_logger])


def log_figures(module: pl.LightningModule, d: Dict[str, Figure]):
    assert isinstance(module, pl.LightningModule)
    module_loggers = (
        module.logger if hasattr(module.logger, "__getitem__") else [module.loggers]
    )
    image_loggers = []
    for lgr in module_loggers:
        if type(lgr) == NeptuneLogger:
            # log_image(log_name, image, step=None)
            image_loggers.append(lgr.log_image)
        elif type(lgr) == TensorBoardLogger:
            # SummaryWriter.add_figure(self, tag, figure)
            image_loggers.append(lgr.experiment.add_figure)
        elif type(lgr) == WandbLogger:
            import wandb  # noqa: F401

            wandb_log = lgr.experiment.log

            def log_figure(tag, fig):
                im = wandb.Image(fig2img(fig), caption=tag)
                return wandb_log({tag: im})

            image_loggers.append(log_figure)
        elif type(lgr) == ResultsLogger:
            image_loggers.append(lgr.log_figure)

    if not image_loggers:
        logger.warn(
            f"Unable to log figures {d.keys()}: No compatible logger found among {module_loggers}"
        )
        return

    for k, v in d.items():
        for log in image_loggers:
            log(k, v)


class ResultsLogger(LightningLoggerBase):
    def __init__(self, prefix="test", save_to: str = None):
        super().__init__()
        self.results = {}
        self.prefix = prefix
        self.log_dir = save_to

    def _fix_name_perfix(self, s: str, replace="test/") -> str:
        if not self.prefix:
            return s

        if s.startswith(replace):
            return f"{self.prefix}/{s[5:]}"
        else:
            return f"{self.prefix}/{s}"

    @property
    def experiment(self):
        return None

    @rank_zero_only
    def log_hyperparams(self, params):
        if self.log_dir:
            dump_yaml(
                path=Path(self.log_dir) / f"{self.prefix}_hparams.yaml",
                data=params,
            )

    @rank_zero_only
    def log_metrics(self, metrics: Dict, step):
        self.results = {self._fix_name_perfix(k): float(v) for k, v in metrics.items()}
        if self.log_dir:
            dump_yaml(
                path=Path(self.log_dir) / f"{self.prefix}_results.yaml",
                data={k: float(v) for k, v in self.results.items()},
            )

    def log_figure(self, tag: str, fig: Figure):
        if self.log_dir:
            fig_path = str(Path(self.log_dir) / f"{self.prefix}_{tag}.png")
            logger.info(f"Saving confusion matrix to {fig_path}")
            fig.savefig(fig_path)

    @rank_zero_only
    def finalize(self, status):
        pass

    @property
    def save_dir(self) -> Optional[str]:
        return self.log_dir

    @property
    def name(self):
        return "ResultsLogger"

    @property
    def version(self):
        return "1"


StepOutputs = List[Dict[str, Any]]


class CheckpointEveryNSteps(pl.Callback):
    """
    Save a checkpoint every N steps, instead of Lightning's default that checkpoints
    based on validation loss.

    Courtesy of Andrew Jong
    https://github.com/PyTorchLightning/pytorch-lightning/issues/2534
    """

    def __init__(
        self,
        save_step_frequency,
        prefix="N-Step-Checkpoint",
        use_modelcheckpoint_filename=False,
    ):
        """
        Args:
            save_step_frequency: how often to save in steps
            prefix: add a prefix to the name, only used if
                use_modelcheckpoint_filename=False
            use_modelcheckpoint_filename: just use the ModelCheckpoint callback's
                default filename, don't use ours.
        """
        self.save_step_frequency = save_step_frequency
        self.prefix = prefix
        self.use_modelcheckpoint_filename = use_modelcheckpoint_filename

    def on_batch_end(self, trainer: pl.Trainer, _):
        """ Check if we should save a checkpoint after every train batch """
        epoch = trainer.current_epoch
        global_step = trainer.global_step
        if global_step % self.save_step_frequency == 0 and global_step != 0:
            if self.use_modelcheckpoint_filename:
                filename = trainer.checkpoint_callback.filename
            else:
                filename = f"{self.prefix}_{epoch}_{global_step}.ckpt"
            ckpt_path = os.path.join(trainer.checkpoint_callback.dirpath, filename)
            pl.utilities.rank_zero_info(f"Saving model to {str(ckpt_path)}")
            trainer.save_checkpoint(ckpt_path)



# from ride.profile import Profileable
import inspect
from abc import ABC
from typing import Any, List, Sequence, Union

import pytorch_lightning as pl
from corider import Configs as _Configs
from pytorch_lightning.utilities.parsing import AttributeDict
from supers import supers
from torch.utils.data import DataLoader

from ride.utils.logging import getLogger
from ride.utils.utils import (
    DictLike,
    attributedict,
    merge_attributedicts,
    missing_or_not_in_other,
    name,
    some,
)

logger = getLogger(__name__)


DataShape = Union[int, Sequence[int], Sequence[Sequence[int]]]


class Configs(_Configs):
    @staticmethod
    def collect(cls: "RideModule") -> "Configs":
        c: Configs = sum([c.configs() for c in cls.__bases__ if hasattr(c, "configs")])  # type: ignore
        return c

    def default_values(self):
        return attributedict({k: v.default for k, v in self.values.items()})


def _init_subsubclass(cls):
    orig_init = cls.__init__

    def init(self, hparams: DictLike, *args, **kwargs):
        super(cls, self).__init__(hparams, *args, **kwargs)
        apply_init_args(orig_init, self, self.hparams, *args, **kwargs)

    cls.__init__ = init


def _init_subclass(cls):
    # Validate inheritance order
    assert (
        cls.__bases__[0] == RideModule or cls.__bases__[0].__bases__[0] == RideModule
    ), """RideModule must come first in inheritance order, e.g.:
    class YourModule(RideModule, OtherMixin):
        ..."""

    if cls.__bases__[0] is not RideModule:
        _init_subsubclass(cls)
        return

    if not cls.__bases__[-1] == pl.LightningModule:
        cls.__bases__ = (*cls.__bases__, pl.LightningModule)

    # Warn if there is no forward
    if missing_or_not_in_other(
        cls, pl.LightningModule, {"forward"}, must_be_callable=True
    ):
        # if not (some(cls, "forward") and callable(cls.forward)):
        logger.warning(
            f"No `forward` function found in {name(cls)}. Did you forget to define it?"
        )

    # Ensure lifecycle
    lifecycle_steps = {
        "training_step",
        "validation_step",
        "test_step",
    }
    missing_lifecycle_steps = missing_or_not_in_other(
        cls, pl.LightningModule, lifecycle_steps
    )
    if missing_lifecycle_steps:
        logger.info(f"Missing lifecycle steps {missing_lifecycle_steps} in {name(cls)}")
        logger.info("🔧 Adding ride.Lifecycle automatically")
        # Import here to break cyclical import
        from ride.lifecycle import Lifecycle

        cls.__bases__ = (
            *cls.__bases__[:-1],
            Lifecycle,
            *cls.__bases__[-1:],
        )

    # Ensure dataset
    dataset_steps = {"train_dataloader", "val_dataloader", "test_dataloader"}
    missing_dataset_steps = missing_or_not_in_other(
        cls, pl.LightningModule, dataset_steps
    )
    if missing_dataset_steps:
        logger.warning(
            f"No dataloader funcions {missing_dataset_steps} found in {name(cls)}"
        )
        logger.info(
            "🔧 Adding ride.RideDataset automatically and assuming that `self.datamodule`, `self.input_shape`, and `self.output_shape` will be provided by user"
        )
        cls.__bases__ = (
            *cls.__bases__[:-1],
            RideDataset,
            *cls.__bases__[-1:],
        )

    # Ensure optimizer
    if missing_or_not_in_other(cls, pl.LightningModule, {"configure_optimizers"}):
        logger.info(f"`configure_optimizers` not found in in {name(cls)}")
        logger.info("🔧 Adding ride.SgdOptimizer automatically")
        # Import here to break cyclical import
        from ride.optimizers import SgdOptimizer

        cls.__bases__ = (
            *cls.__bases__[:-1],
            SgdOptimizer,
            *cls.__bases__[-1:],
        )

    # Monkeypatch derived module init
    orig_init = cls.__init__

    def init(self, hparams: DictLike = {}, *args, **kwargs):
        pl.LightningModule.__init__(self)
        self.hparams = merge_attributedicts(self.configs().default_values(), hparams)
        sself = (
            self
            if type(self).__bases__[0] == RideModule
            else supers(self)._superclasses[0]
        )
        supers(sself)[1:-1].__init__(self.hparams)
        apply_init_args(orig_init, self, self.hparams, *args, **kwargs)
        supers(sself).on_init_end(self.hparams, *args, **kwargs)
        supers(self).validate_attributes()

    cls.__init__ = init

    # Monkeypatch derived module configs
    orig_configs = getattr(cls, "configs", None)

    @staticmethod
    def configs():
        c = Configs.collect(cls)
        if orig_configs:
            c += orig_configs()
        return c

    cls.configs = configs


def apply_init_args(fn, self, hparams, *args, **kwargs):
    spec = inspect.getfullargspec(fn)
    valid_kwargs = (
        kwargs
        if spec.varkw == "kwargs"
        else {k: v for k, v in kwargs.items() if k in spec.args}
    )
    if len(spec.args) == 1:
        return fn(self)
    else:
        return fn(self, hparams, *args, **valid_kwargs)


class RideModule:
    """
    Base-class for modules using the Ride ecosystem.

    This module should be inherited as the highest-priority parent.

    Example::

        class MyModule(ride.RideModule, ride.SgdOneCycleOptimizer):
            def __init__(self, hparams):
                ...

    It handles proper initialisation of `RideMixin` parents and adds automatic attribute validation.

    If `pytorch_lightning.LightningModule` is omitted as lowest-priority parent, `RideModule` will automatically add it.

    If `training_step`, `validation_step`, and `test_step` methods are not found, the `ride.Lifecycle` will be automatically mixed in by this module.
    """

    def __init_subclass__(cls):
        _init_subclass(cls)

    @property
    def hparams(self) -> AttributeDict:
        if not hasattr(self, "_hparams"):
            self._hparams = AttributeDict()
        return self._hparams

    @hparams.setter
    def hparams(self, hp: Union[dict, AttributeDict, Any]):
        # Overload the version in pytorch_lightning core to omit DeprecationWarning
        self._hparams = attributedict(hp)

    @classmethod
    def with_dataset(cls, ds: "RideDataset"):
        DerivedRideModule = type(
            f"{name(cls)}With{name(ds)}", cls.__bases__, dict(cls.__dict__)
        )

        new_bases = [b for b in cls.__bases__ if not issubclass(b, RideDataset)]
        old_dataset = [b for b in cls.__bases__ if issubclass(b, RideDataset)]
        assert len(old_dataset) <= 1, "`RideModule` should only have one `RideDataset`"
        if old_dataset and issubclass(old_dataset[0], RideClassificationDataset):
            assert issubclass(
                ds, RideClassificationDataset
            ), "A `RideClassificationDataset` should be replaced by a `RideClassificationDataset`"
        new_bases.insert(-1, ds)
        DerivedRideModule.__bases__ = tuple(new_bases)

        return DerivedRideModule


class RideMixin(ABC):
    def __init__(self, hparams: AttributeDict, *args, **kwargs):
        ...

    def on_init_end(self, hparams: AttributeDict, *args, **kwargs):
        ...

    def validate_attributes(self):
        ...


class OptimizerMixin(RideMixin):
    ...


class RideDataset(RideMixin):
    input_shape: DataShape
    output_shape: DataShape

    def validate_attributes(self):
        assert type(getattr(self, "input_shape", None)) in {
            int,
            list,
            tuple,
        }, "RideDataset should define `input_shape` but none was found."
        assert type(getattr(self, "output_shape", None)) in {
            int,
            list,
            tuple,
        }, "RideDataset should define `output_shape` but none was found."

        for n in RideDataset.configs().names:
            assert some(
                self, f"hparams.{n}"
            ), "`self.hparams.{n}` not found in Dataset. Did you forget to include its `configs`?"

    @staticmethod
    def configs() -> Configs:
        c = Configs()
        c.add(
            name="batch_size",
            type=int,
            default=16,
            strategy="constant",
            description="Batch size for dataset.",
        )
        c.add(
            name="num_workers",
            type=int,
            default=0,
            strategy="constant",
            description="Number of workers in dataloader.",
        )
        return c

    def train_dataloader(self, *args: Any, **kwargs: Any) -> DataLoader:
        """ The train dataloader """
        assert some(
            self, "datamodule.train_dataloader"
        ), f"{name(self)} should either have a `self.datamodule: pl.LightningDataModule` or overload the `train_dataloader` function."
        return self.datamodule.train_dataloader

    def val_dataloader(
        self, *args: Any, **kwargs: Any
    ) -> Union[DataLoader, List[DataLoader]]:
        """ The val dataloader """
        assert some(
            self, "datamodule.val_dataloader"
        ), f"{name(self)} should either have a `self.datamodule: pl.LightningDataModule` or overload the `val_dataloader` function."
        return self.datamodule.val_dataloader

    def test_dataloader(
        self, *args: Any, **kwargs: Any
    ) -> Union[DataLoader, List[DataLoader]]:
        """ The test dataloader """
        assert some(
            self, "datamodule.test_dataloader"
        ), f"{name(self)} should either have a `self.datamodule: pl.LightningDataModule` or overload the `test_dataloader` function."
        return self.datamodule.test_dataloader


class RideClassificationDataset(RideDataset):
    classes: List[str]

    @property
    def num_classes(self) -> int:
        return len(self.classes)

    def validate_attributes(self):
        RideDataset.validate_attributes(self)
        assert type(getattr(self, "classes", None)) in {
            list,
            tuple,
        }, "Ride RideClassificationDataset should define `classes` but none was found."



from ride.core import AttributeDict, RideClassificationDataset, Configs
from ride.utils.env import DATASETS_PATH


class MnistDataset(RideClassificationDataset):
    """
    Example Mnist Dataset
    Modified from https://github.com/PyTorchLightning/pytorch-lightning/blob/master/pl_examples/basic_examples/mnist_datamodule.py
    """

    @staticmethod
    def configs():
        c = Configs.collect(MnistDataset)
        c.add(
            name="val_split",
            type=int,
            default=5000,
            strategy="constant",
            description="Number samples from train dataset used for validation split.",
        )
        c.add(
            name="normalize",
            type=int,
            default=1,
            choices=[0, 1],
            strategy="constant",
            description="Whether to normalize dataset.",
        )
        return c

    def __init__(self, hparams: AttributeDict):
        import pl_bolts

        self.datamodule = pl_bolts.datamodules.MNISTDataModule(
            data_dir=DATASETS_PATH,
            val_split=self.hparams.val_split,
            num_workers=self.hparams.num_workers,
            normalize=self.hparams.normalize,
            batch_size=self.hparams.batch_size,
            seed=42,
            shuffle=True,
            pin_memory=self.hparams.num_workers > 1,
            drop_last=False,
        )
        self.output_shape = 10
        self.classes = list(range(10))
        self.input_shape = self.datamodule.dims



import functools
import inspect
import math
import re
from argparse import Namespace
from contextlib import contextmanager
from functools import wraps
from operator import attrgetter
from typing import Any, Callable, Collection, Dict, Set, Union

from pytorch_lightning.utilities.parsing import AttributeDict

DictLike = Union[AttributeDict, Dict[str, Any], Namespace]


def once(fn: Callable):
    mem = set()

    @wraps(fn)
    def wrapped(*args, **kwargs):
        h = hash((args, str(kwargs)))
        if h in mem:
            return
        else:
            mem.add(h)
            return fn(*args, **kwargs)

    return wrapped


def rsetattr(obj, attr, val):
    pre, _, post = attr.rpartition(".")
    return setattr(rgetattr(obj, pre) if pre else obj, post, val)


def rgetattr(obj, attr, *args):
    def _getattr(obj, attr):
        return getattr(obj, attr, *args)

    return functools.reduce(_getattr, [obj] + attr.split("."))


def attributedict(dict_like: DictLike) -> AttributeDict:
    """If given a dict, it is converted it to an argparse.AttributeDict. Otherwise, no change is made"""
    if isinstance(dict_like, AttributeDict):
        return dict_like
    elif isinstance(dict_like, Namespace):
        return AttributeDict(vars(dict_like))
    elif isinstance(dict_like, dict):
        return AttributeDict(**dict_like)
    else:
        raise ValueError(f"Unable to convert type {type(dict_like)} to AttributeDict")


def to_dict(d):
    if type(d) == Namespace:
        return vars(d)
    else:
        return dict(d)


def merge_dicts(*args):
    if len(args) == 0:
        return {}
    if len(args) == 1:
        return args[0]
    acc = to_dict(args[0])
    for a in args[1:]:
        acc = {**acc, **to_dict(a)}
    return acc


def merge_attributedicts(*args):
    return attributedict(merge_dicts(*args))


def some(self, attr: str):
    try:
        a = attrgetter(attr)(self)
        return a is not None
    except Exception:
        return False


def some_callable(self, attr: str, min_num_args=0, max_num_args=math.inf):
    try:
        fn = attrgetter(attr)(self)
        if not callable(fn):
            return False
        num_args = len(inspect.getfullargspec(fn).args)
        return min_num_args <= num_args and num_args <= max_num_args
    except Exception:
        return False


def get(self, attr: str):
    try:
        a = attrgetter(attr)(self)
        return a
    except KeyError:
        return None


def differ_and_exist(a, b):
    return a is b and a is not None


def missing(self, attrs: Collection[str]) -> Set[str]:
    return {a for a in attrs if not some(self, a)}


def missing_or_not_in_other(
    first, other, attrs: Collection[str], must_be_callable=False
) -> Set[str]:
    some_ = some_callable if must_be_callable else some
    return {
        a
        for a in attrs
        if not some_(first, a) or differ_and_exist(get(first, a), get(other, a))
    }


def name(any):
    if hasattr(any, "__name__"):
        return any.__name__
    else:
        return any.__class__.__name__


def prefix_keys(prefix: str, dictionary: Dict) -> Dict:
    return {f"{prefix}{k}": v for k, v in dictionary.items()}


def camel_to_snake(s: str) -> str:
    """Convert from camel-case to snake-case
    Source: https://stackoverflow.com/questions/1175208/elegant-python-function-to-convert-camelcase-to-snake-case
    """
    s = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", s)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s).lower()


@contextmanager
def temporary_parameter(obj, attr, val):
    prev_val = rgetattr(obj, attr)
    rsetattr(obj, attr, val)
    yield obj
    rsetattr(obj, attr, prev_val)



from ride import hparamsearch  # noqa: F401


def test_add_argparse_args():
    assert False


def test_run():
    assert False


def test_call():
    assert False


def test_dump():
    assert False


def test_load():
    assert False



from typing import List, Optional, Union

from pytorch_lightning.utilities import device_parser


def parse_gpus(args_gpus: Optional[Union[int, str, List[int]]]) -> List[int]:
    return device_parser.parse_gpu_ids(args_gpus) or []


def parse_num_gpus(args_gpus: Optional[Union[int, str, List[int]]]) -> int:
    return len(parse_gpus(args_gpus))



from operator import attrgetter
from typing import Dict, Sequence

import torch

from ride.core import Configs, RideMixin
from ride.utils.logging import getLogger

logger = getLogger(__name__)


class Unfreezable(RideMixin):
    hparams: ...

    def validate_attributes(self):
        for hparam in self.configs().names:
            attrgetter(f"hparams.{hparam}")(self)
        attrgetter("hparams.max_epochs")(self)

    @staticmethod
    def configs() -> Configs:
        c = Configs()

        c.add(
            name="unfreeze_from_epoch",
            type=int,
            default=-1,
            description="Number of epochs to wait before starting gradual unfreeze. If -1, unfreeze is omitted.",
        )
        c.add(
            name="unfreeze_layers_must_include",
            type=str,
            default="",
            description="String that must be contained in layer names which should be unfrozen. If empty, this feature is disabled.",
        )
        c.add(
            name="unfreeze_epoch_step",
            type=int,
            default=1,
            description="Number of epochs to train before next unfreeze.",
        )
        c.add(
            name="unfreeze_layers_initial",
            type=int,
            default=1,
            strategy="choice",
            description="Number layers to unfreeze initially. If `-1`, it will be equal to total_layers",
        )
        c.add(
            name="unfreeze_layer_step",
            type=int,
            default=1,
            description="Number additional layers to unfreeze at each unfreeze step.",
        )
        c.add(
            name="unfreeze_layers_max",
            type=int,
            default=-1,
            description="Maximum number of layers to unfreeze. If `-1`, it will be equal to total_layers",
        )
        return c

    def on_init_end(
        self,
        hparams,
        layers_to_unfreeze: Sequence[torch.nn.Module] = None,
        *args,
        **kwargs,
    ):
        self.layers_to_unfreeze = (
            layers_to_unfreeze
            if layers_to_unfreeze is not None
            else get_modules_to_unfreeze(
                self, self.hparams.unfreeze_layers_must_include
            )
        )

        # Gradual unfreeze linear schedule
        self.unfreeze_schedule = (
            linear_unfreeze_schedule(
                initial_epoch=self.hparams.unfreeze_from_epoch,
                total_layers=len(self.layers_to_unfreeze),
                step_size=self.hparams.unfreeze_layer_step,
                init_layers=self.hparams.unfreeze_layers_initial,
                max_layers=self.hparams.unfreeze_layers_max,
                epoch_step=self.hparams.unfreeze_epoch_step,
            )
            if self.hparams.unfreeze_from_epoch > -1
            else {}
        )

    def on_traning_epoch_start(self, epoch: int):
        # Called by TrainValTestStepsMixin
        if epoch in self.unfreeze_schedule:
            num_layers = self.unfreeze_schedule[epoch]
            logger.info(f"Epoch {epoch}: Unfreezing {num_layers} layer(s)")
            unfreeze_from_end(self.layers_to_unfreeze, num_layers)


def get_modules_to_unfreeze(
    parent_module: torch.nn.Module, name_must_include=""
) -> Sequence[torch.nn.Module]:
    return [
        m
        for (n, m) in parent_module.named_modules()
        if (name_must_include in n) and (hasattr(m, "weight") or hasattr(m, "bias"))
    ]


def unfreeze_from_end(
    layers: Sequence[torch.nn.Module],
    num_layers_from_end: int,
    freeze_others=True,
):
    if freeze_others:
        for layer in layers[:-num_layers_from_end:]:
            for param in layer.parameters():
                param.requires_grad = False

    num_unfrozen = 0
    layer_iter = iter(layers[::-1])
    while num_unfrozen < num_layers_from_end:
        try:
            layer = next(layer_iter)
            for param in layer.parameters():
                param.requires_grad = True
            num_unfrozen += 1
        except StopIteration:
            break


def linear_unfreeze_schedule(
    initial_epoch: int,
    total_layers: int,
    step_size: int = 1,
    init_layers: int = 0,
    max_layers: int = -1,
    epoch_step: int = 1,
) -> Dict[int, int]:
    if init_layers == -1:
        init_layers = total_layers
    if max_layers == -1:
        max_layers = total_layers
    if step_size == -1:
        step_size = total_layers - init_layers
    return {
        **{
            epoch * epoch_step + initial_epoch: i * step_size + init_layers
            for i, epoch in enumerate(range(total_layers * step_size))
            if i * step_size + init_layers <= max_layers
        },
        **{0: init_layers},
    }



from enum import Enum
from operator import attrgetter
from typing import Dict, List

import matplotlib.pyplot as plt
import numpy as np
import pytorch_lightning as pl
import torch
from ptflops import get_model_complexity_info
from supers import supers
from torch import Tensor

from ride.core import Configs, RideMixin
from ride.utils.logging import getLogger
from ride.utils.utils import merge_dicts, name

MetricDict = Dict[str, Tensor]
StepOutputs = List[Dict[str, Tensor]]

logger = getLogger(__name__)


class OptimisationDirection(Enum):
    MIN = "min"
    MAX = "max"


class MetricMixin(RideMixin):
    """Abstract base class for Ride modules"""

    def __init_subclass__(cls):
        if not hasattr(cls, "_metrics"):
            logger.error(
                f"Subclasses of MetricMixin should define a `_metrics` classmethod, but none was found in {name(cls)}"
            )

    @classmethod
    def metrics(cls) -> Dict[str, str]:
        ms = merge_dicts(
            *[c._metrics() for c in cls.__bases__ if issubclass(c, MetricMixin)]
        )
        return ms

    @classmethod
    def metric_names(cls) -> List[str]:
        return list(sorted(cls.metrics().keys()))

    def metrics_step(self, *args, **kwargs) -> MetricDict:
        return {}

    def metrics_epoch(
        self, outputs: StepOutputs, prefix: str = "", **kwargs
    ) -> MetricDict:
        return {}

    def collect_metrics(self, preds: Tensor, targets: Tensor) -> MetricDict:
        device = preds.device
        mdlist: List[MetricDict] = supers(self).metrics_step(preds, targets)  # type: ignore
        return {
            k: v.to(device=device) if hasattr(v, "to") else v
            for md in mdlist
            for k, v in md.items()
        }

    def collect_epoch_metrics(self, preds: Tensor, targets: Tensor) -> MetricDict:
        device = preds.device
        mdlist: List[MetricDict] = supers(self).metrics_epoch(preds, targets)  # type: ignore
        return {
            k: v.to(device=device) if hasattr(v, "to") else v
            for md in mdlist
            for k, v in md.items()
        }


class MeanAveragePrecisionMetric(MetricMixin):
    """Mean Average Precision (mAP) metric"""

    def validate_attributes(self):
        for attribute in ["hparams.loss", "classes"]:
            attrgetter(attribute)(self)

    @classmethod
    def _metrics(cls):
        return {"mAP": OptimisationDirection.MAX}

    def metrics_step(self, preds: Tensor, targets: Tensor, **kwargs) -> MetricDict:
        map = torch.tensor(-1.0)
        try:
            map = compute_map(self)(preds, targets)
        except RuntimeError:
            logger.error("Unable to compute mAP.")
        return {"mAP": map}

    def metrics_epoch(self, preds: Tensor, targets: Tensor, **kwargs) -> MetricDict:
        preds = preds.detach().cpu().numpy()
        targets = targets.detach().cpu().numpy()
        map = torch.tensor(-1.0)
        try:
            map, wmap, _ = charades_mean_average_precision(preds, targets)
        except RuntimeError:
            logger.error("Unable to compute mAP.")
        return {"mAP": torch.tensor(map)}


def compute_map(self):
    if "map_fn" not in vars():
        if "binary" in self.hparams.loss:
            map_fn = pl.metrics.classification.AveragePrecision(pos_label=1)
        else:
            map_fn = pl.metrics.classification.AveragePrecision(
                num_classes=len(self.classes)
            )
    return map_fn


def charades_mean_average_precision(preds, targets):
    """
    Returns mAP, weighted mAP, and AP array
    Source: https://github.com/gsig/charades-algorithms
    """
    preds = preds.copy()
    empty = np.sum(targets, axis=1) == 0
    preds[empty, :] = np.NINF

    m_aps = []
    n_classes = preds.shape[1]
    for oc_i in range(n_classes):
        sorted_idxs = np.argsort(-preds[:, oc_i])
        tp = targets[:, oc_i][sorted_idxs] == 1
        fp = np.invert(tp)
        n_pos = tp.sum()
        if n_pos < 0.1:
            m_aps.append(float("nan"))
            continue
        # fp.sum()
        f_pcs = np.cumsum(fp)
        t_pcs = np.cumsum(tp)
        prec = t_pcs / (f_pcs + t_pcs).astype(float)
        avg_prec = 0
        for i in range(preds.shape[0]):
            if tp[i]:
                avg_prec += prec[i]
        m_aps.append(avg_prec / n_pos.astype(float))
    m_aps = np.array(m_aps)
    m_ap = np.nanmean(m_aps)  # ignore nan in calculation of mean
    w_ap = m_aps * targets.sum(axis=0) / targets.sum().sum().astype(float)
    return m_ap, w_ap, m_aps


def TopKAccuracyMetric(*Ks) -> MetricMixin:
    if not Ks:
        Ks = [1, 3, 5]

    for k in Ks:
        assert type(k) == int and k > 0

    class TopKAccuracyMetricClass(MetricMixin):
        """Top K accuracy metrics: top1acc, top3acc, top5acc"""

        @classmethod
        def _metrics(cls):
            return {f"top{k}acc": OptimisationDirection.MAX for k in Ks}

        def metrics_step(self, preds: Tensor, targets: Tensor, **kwargs) -> MetricDict:
            ks = [k for k in Ks]
            accs = [torch.tensor(-1.0) for _ in ks]
            try:
                accs = topk_accuracies(preds, targets, ks)
            except RuntimeError:
                logger.error("Unable to compute top-k accuracy.")
            return {f"top{k}acc": accs[i] for i, k in enumerate(ks)}

    return TopKAccuracyMetricClass


class FlopsMetric(MetricMixin):
    """Computes Floating Point Operations (FLOPs) for the model and adds it as metric"""

    @classmethod
    def _metrics(cls):
        return {"flops": OptimisationDirection.MIN}

    def __init__(self, *args, **kwargs) -> None:
        assert isinstance(self, torch.nn.Module)
        self.flops = flops(self)  # type: ignore

    def metrics_step(self, preds: Tensor, targets: Tensor, **kwargs) -> MetricDict:
        return {"flops": torch.tensor(self.flops)}


class FlopsWeightedAccuracyMetric(FlopsMetric):
    """Computes acc * (flops / target_gflops) ** (-0.07)"""

    @classmethod
    def _metrics(cls):
        return {"flops_weighted_acc": OptimisationDirection.MAX}

    def __init__(self, *args, **kwargs):
        FlopsMetric.__init__(self, *args, **kwargs)

    def validate_attributes(self):
        for hparam in FlopsWeightedAccuracyMetric.configs().names:
            attrgetter(f"hparams.{hparam}")(self)

    @staticmethod
    def configs() -> Configs:
        c = Configs()
        c.add(
            name="target_gflops",
            type=float,
            default=2.0,
            strategy="constant",
            description="Target (Giga) Floating Point Operations per Second.",
        )
        return c

    def metrics_step(self, preds: Tensor, targets: Tensor, **kwargs) -> MetricDict:
        acc = topk_accuracies(preds, targets, ks=[1])[0]
        return {
            **FlopsMetric.metrics_step(self, preds, targets, **kwargs),
            "flops_weighted_acc": acc
            * (self.flops / self.hparams.target_gflops) ** (-0.07),
        }


def topks_correct(preds: Tensor, labels: Tensor, ks: List[int]) -> List[Tensor]:
    """
    Given the predictions, labels, and a list of top-k values, compute the
    number of correct predictions for each top-k value.

    Args:
        preds (array): array of predictions. Dimension is batchsize
            N x ClassNum.
        labels (array): array of labels. Dimension is batchsize N.
        ks (list): list of top-k values. For example, ks = [1, 5] correspods
            to top-1 and top-5.

    Returns:
        topks_correct (list): list of numbers, where the `i`-th entry
            corresponds to the number of top-`ks[i]` correct predictions.
    """
    assert preds.size(0) == labels.size(
        0
    ), "Batch dim of predictions and labels must match"

    max_k = int(preds.shape[-1])

    # Find the top max_k predictions for each sample
    _, top_max_k_inds = torch.topk(preds, max_k, dim=1, largest=True, sorted=True)
    # (batch_size, max_k) -> (max_k, batch_size).
    top_max_k_inds = top_max_k_inds.t()
    # (batch_size, ) -> (max_k, batch_size).
    rep_max_k_labels = labels.view(1, -1).expand_as(top_max_k_inds)
    # (i, j) = 1 if top i-th prediction for the j-th sample is correct.
    top_max_k_correct = top_max_k_inds.eq(rep_max_k_labels)
    # Compute the number of topk correct predictions for each k.
    topks_correct: List[Tensor] = [
        top_max_k_correct[: min(k, max_k), :].reshape(-1).float().sum() for k in ks
    ]
    return topks_correct


def topk_errors(preds: Tensor, labels: Tensor, ks: List[int]):
    """
    Computes the top-k error for each k.
    Args:
        preds (array): array of predictions. Dimension is N.
        labels (array): array of labels. Dimension is N.
        ks (list): list of ks to calculate the top accuracies.
    """
    num_topks_correct = topks_correct(preds, labels, ks)
    return [(1.0 - x / preds.size(0)) for x in num_topks_correct]


def topk_accuracies(preds: Tensor, labels: Tensor, ks: List[int]):
    """
    Computes the top-k accuracy for each k.
    Args:
        preds (array): array of predictions. Dimension is N.
        labels (array): array of labels. Dimension is N.
        ks (list): list of ks to calculate the top accuracies.
    """
    num_topks_correct = topks_correct(preds, labels, ks)
    return [(x / preds.size(0)) for x in num_topks_correct]


def flops(model: torch.nn.Module):
    """Compute the Floating Point Operations per Second for the model"""
    return get_model_complexity_info(
        model,
        model.input_shape,
        as_strings=False,
        print_per_layer_stat=True,
        verbose=True,
    )[0]


def params_count(model: torch.nn.Module):
    """
    Compute the number of parameters.
    Args:
        model (model): model to count the number of parameters.
    """
    return np.sum([p.numel() for p in model.parameters()])


def make_confusion_matrix(  # noqa: C901
    preds,
    targets,
    num_classes,
    group_names=None,
    categories="auto",
    count=True,
    percent=True,
    cbar=True,
    xyticks=True,
    xyplotlabels=True,
    sum_stats=False,
    figsize=None,
    cmap="Blues",
    title=None,
    save_as=None,
):
    """
    Modified from https://github.com/DTrimarchi10/confusion_matrix
    This function will make a pretty plot of an sklearn Confusion Matrix cm using a Seaborn heatmap visualization.

    Arguments
    ---------
    preds:         Predictions

    targets:       Targets

    group_names:   List of strings that represent the labels row by row to be shown in each square.

    categories:    List of strings containing the categories to be displayed on the x,y axis. Default is 'auto'

    count:         If True, show the raw number in the confusion matrix. Default is True.

    percent:     If True, show the proportions for each category. Default is True.

    cbar:          If True, show the color bar. The cbar values are based off the values in the confusion matrix.
                   Default is True.

    xyticks:       If True, show x and y ticks. Default is True.

    xyplotlabels:  If True, show 'True Label' and 'Predicted Label' on the figure. Default is True.

    sum_stats:     If True, display summary statistics below the figure. Default is True.

    figsize:       Tuple representing the figure size. Default will be the matplotlib rcParams value.

    cmap:          Colormap of the values displayed from matplotlib.pyplot.cm. Default is 'Blues'
                   See http://matplotlib.org/examples/color/colormaps_reference.html

    title:         Title for the heatmap. Default is None.

    """

    from pytorch_lightning.metrics.functional.confusion_matrix import confusion_matrix
    from seaborn import heatmap

    font_logger = getLogger("matplotlib.font_manager")
    font_logger.propagate = False

    if len(preds.shape) > 1:
        preds = preds.squeeze().argmax(-1)
    assert targets.shape and preds.shape and len(targets.shape) == 1
    cf = confusion_matrix(preds, targets, num_classes).cpu().numpy()

    # CODE TO GENERATE TEXT INSIDE EACH SQUARE
    blanks = ["" for i in range(cf.size)]

    if group_names and len(group_names) == cf.size:
        group_labels = ["{}\n".format(value) for value in group_names]
    else:
        group_labels = blanks

    if count:
        group_counts = ["{0:0.0f}\n".format(value) for value in cf.flatten()]
    else:
        group_counts = blanks

    if percent:
        group_percentages = [
            "{0:.2%}".format(value) for value in cf.flatten() / np.sum(cf)
        ]
    else:
        group_percentages = blanks

    box_labels = [
        f"{v1}{v2}{v3}".strip()
        for v1, v2, v3 in zip(group_labels, group_counts, group_percentages)
    ]
    box_labels = np.asarray(box_labels).reshape(cf.shape[0], cf.shape[1])

    # CODE TO GENERATE SUMMARY STATISTICS & TEXT FOR SUMMARY STATS
    if sum_stats:
        # Accuracy is sum of diagonal divided by total observations
        accuracy = np.trace(cf) / float(np.sum(cf))

        # if it is a binary confusion matrix, show some more stats
        if len(cf) == 2:
            # Metrics for Binary Confusion Matrices
            precision = cf[1, 1] / sum(cf[:, 1])
            recall = cf[1, 1] / sum(cf[1, :])
            f1_score = 2 * precision * recall / (precision + recall)
            stats_text = "\n\nAccuracy={:0.3f}\nPrecision={:0.3f}\nRecall={:0.3f}\nF1 Score={:0.3f}".format(
                accuracy, precision, recall, f1_score
            )
        else:
            stats_text = "\n\nAccuracy={:0.3f}".format(accuracy)
    else:
        stats_text = ""

    # SET FIGURE PARAMETERS ACCORDING TO OTHER ARGUMENTS
    if figsize is None:
        # Get default figure size if not set
        figsize = plt.rcParams.get("figure.figsize")

    if xyticks is False:
        # Do not show categories if xyticks is False
        categories = False

    # MAKE THE HEATMAP VISUALIZATION
    fig = plt.figure(figsize=figsize)
    heatmap(
        cf,
        annot=box_labels,
        fmt="",
        cmap=cmap,
        cbar=cbar,
        xticklabels=categories,
        yticklabels=categories,
    )

    if xyplotlabels:
        plt.ylabel("True label")
        plt.xlabel("Predicted label" + stats_text)
    else:
        plt.xlabel(stats_text)

    if title:
        plt.title(title)

    return fig



import contextlib
from abc import abstractmethod
from time import time
from typing import Any, Dict, Tuple, overload

import numpy as np
import torch
from torch.autograd.profiler import format_time
from tqdm import tqdm

from ride.utils.gpus import parse_gpus, parse_num_gpus
from ride.utils.logging import getLogger
from ride.utils.utils import name, some

logger = getLogger(__name__)


@overload
def profile(model: torch.nn.Module, detailed=False) -> float:
    ...  # pragma: no cover


@overload
def profile(
    model: torch.nn.Module, detailed=True
) -> Tuple[float, torch.autograd.profiler.EventList,]:
    ...  # pragma: no cover


def profile(model: torch.nn.Module, detailed=True):
    for attr in [
        "device",
        "input_shape",
        "hparams.batch_size",
        "hparams.gpus",
        "forward",
    ]:
        assert some(
            model, attr
        ), f"{name(model)} should define `{attr}` but none was found."

    prevously_training = model.training
    model.eval()

    # Move model to GPU if available
    prev_device = model.device
    gpus = parse_gpus(model.hparams.gpus)
    new_device = f"cuda:{gpus[0]}" if gpus else "cpu"
    model.to(device=new_device)

    # Initialise data on CPU
    data = torch.randn(model.hparams.batch_size, *model.input_shape, device="cpu")

    try:
        with torch.no_grad(), torch.autograd.profiler.profile(
            use_cuda=(model.device.type != "cpu"), profile_memory=True
        ) if detailed else contextlib.suppress() as prof:
            # Transfer data from CPU to GPU, compute on GPU, and transfer back to CPU
            start = time()
            model.forward(data.to(device=new_device)).to(device="cpu")
            stop = time()
    except Exception as e:
        raise ValueError(
            f"Caught exception while attempting to profile model. "
            "Did you ensure that you model has a correct `input_shape` attribute (omitting the batch_size dimension)? "
            f"Exception contents: {e}"
        )
    finally:
        model.to(device=prev_device)
        if prevously_training:
            model.train()

    timing = stop - start  # seconds

    if detailed:
        return timing, prof
    else:
        return timing


def compute_num_runs(total_wait_time, single_run_time):
    additional_runs = 1
    single_run_time = max(single_run_time, 1e-7)
    for option in [10000, 1000, 100, 10, 5, 3]:
        if (total_wait_time / single_run_time / option) > 1:
            additional_runs = option
            break
    return additional_runs


def profile_repeatedly(
    model: torch.nn.Module, max_wait_seconds=30, num_runs=None
) -> Tuple[Dict[str, str], torch.autograd.profiler.EventList]:
    for attr in [
        "hparams.batch_size",
        "hparams.gpus",
    ]:
        assert some(
            model, attr
        ), f"{name(model)} should define `{attr}` but none was found."

    # Make initial run to gauge its time. Also serves as a "dry-run"
    single_run_seconds, single_run_prof = profile(model, detailed=True)

    num_runs = (
        num_runs if num_runs else compute_num_runs(max_wait_seconds, single_run_seconds)
    )

    times_s = []
    for _ in tqdm(range(num_runs), desc="Profiling"):
        times_s.append(profile(model, detailed=False))

    batch_size = model.hparams.batch_size
    s_per_sample = np.array(times_s) / batch_size
    samples_per_s = 1 / s_per_sample
    µs_per_sample = s_per_sample * 1e6

    results_dict = {}
    results_dict[
        "time_per_sample"
    ] = f"{format_time(µs_per_sample.mean())} +/- {format_time(µs_per_sample.std())} [{format_time(µs_per_sample.min())}, {format_time(µs_per_sample.max())}]"
    results_dict[
        "samples_per_second"
    ] = f"{samples_per_s.mean():.3f} +/- {samples_per_s.std():.3f} [{samples_per_s.min():.3f}, {samples_per_s.max():.3f}]"
    results_dict["num_runs"] = num_runs
    results_dict["on_gpu"] = parse_num_gpus(model.hparams.gpus) > 0
    results_dict["batch_size"] = batch_size

    return results_dict, single_run_prof


class ProfileableDataset:
    @abstractmethod
    def profile(self) -> Dict[str, Any]:
        ...  # pragma: no cover



from pathlib import Path

import pytest

from examples.simple_classifier import SimpleClassifier

from .dummy_dataset import DummyClassificationDataLoader

from ride import Main  # noqa: F401  # isort:skip


# This test has a dependency on MNIST downloads, which are unstable
@pytest.mark.skip_cicd
def test_simple_classifier():
    m = Main(SimpleClassifier)
    parser = m.argparse(run=False)
    args, _ = parser.parse_known_args()
    args.train = True
    args.test = True
    args.max_epochs = 1
    args.limit_train_batches = 100
    args.limit_val_batches = 10
    args.limit_test_batches = 10
    args.test_confusion_matrix = True
    args.logging_backend = "wandb"
    m.main(args)

    assert "loss" in m.runner.trainer.model.metrics()
    assert "top1acc" in m.runner.trainer.model.metrics()
    assert "top3acc" in m.runner.trainer.model.metrics()
    assert (Path(m.log_dir) / "test_confusion_matrix.png").is_file()


def test_simpler_classifier():
    SimplerClassifier = SimpleClassifier.with_dataset(DummyClassificationDataLoader)
    m = Main(SimplerClassifier)
    parser = m.argparse(run=False)
    args, _ = parser.parse_known_args()
    args.train = True
    args.test = True
    args.max_epochs = 1
    args.limit_train_batches = 100
    args.limit_val_batches = 10
    args.limit_test_batches = 10
    args.test_confusion_matrix = True
    args.loss = "cross_entropy"
    m.main(args)

    assert "loss" in m.runner.trainer.model.metrics()
    assert "top1acc" in m.runner.trainer.model.metrics()
    assert "top3acc" in m.runner.trainer.model.metrics()
    assert (Path(m.log_dir) / "test_confusion_matrix.png").is_file()



""" Dummy Dataset
Modified from
https://github.com/PyTorchLightning/PyTorch-Lightning-Bolts/blob/master/pl_bolts/datamodules/dummy_dataset.py#L5-L42
"""

from typing import Sequence

import torch
from torch.utils.data import DataLoader, Dataset

from ride.core import RideClassificationDataset, Configs
from ride.core import RideDataset
from ride.utils.utils import some


class MeanVarDataset(Dataset):
    def __init__(self, input_shape: Sequence[int], num_samples=10000):
        super().__init__()
        self.input_shape = input_shape
        self.num_samples = num_samples
        self.output_shape = (2,)

    def __len__(self):
        return self.num_samples

    def __getitem__(self, idx):
        x = torch.rand(self.input_shape)
        y = torch.stack(
            [
                x.mean(),
                x.var(),
            ]
        )
        sample = [x, y]
        return sample


class DummyRegressionDataLoader(RideDataset):
    @staticmethod
    def configs() -> Configs:
        c = Configs.collect(DummyRegressionDataLoader)
        c.add(
            name="input_shape",
            type=int,
            default=10,
            strategy="constant",
            description="Input shape for data.",
        )
        return c

    def validate_attributes(self):
        RideDataset.validate_attributes(self)
        for attr in DummyRegressionDataLoader.configs().names:
            assert some(
                self, f"hparams.{attr}"
            ), f"ClassificationLifecycle should define `hparams.{attr}` but none was found."

    def __init__(self, hparams):
        num_workers = 1
        self.input_shape = (hparams.input_shape,)
        self.output_shape = 2

        self._train_dataloader = DataLoader(
            MeanVarDataset(self.input_shape, num_samples=65),
            batch_size=hparams.batch_size,
            num_workers=num_workers,
            shuffle=True,
            pin_memory=num_workers > 1,
        )
        self._val_dataloader = DataLoader(
            MeanVarDataset(self.input_shape, num_samples=4),
            batch_size=hparams.batch_size,
            num_workers=num_workers,
            shuffle=False,
            pin_memory=num_workers > 1,
        )
        self._test_dataloader = DataLoader(
            MeanVarDataset(self.input_shape, num_samples=10),
            batch_size=hparams.batch_size,
            num_workers=num_workers,
            shuffle=False,
            pin_memory=num_workers > 1,
        )

    def train_dataloader(self) -> DataLoader:
        return self._train_dataloader

    def val_dataloader(self) -> DataLoader:
        return self._val_dataloader

    def test_dataloader(self) -> DataLoader:
        return self._test_dataloader


class OverOneDataset(Dataset):
    def __init__(self, input_shape: Sequence[int], num_samples=10000):
        super().__init__()
        self.input_shape = input_shape
        self.num_samples = num_samples
        self.output_shape = (2,)
        self.classes = ["0", "1"]

    def __len__(self):
        return self.num_samples

    def __getitem__(self, idx):
        x = 2 * torch.rand(self.input_shape)
        m = x.mean()
        y = (m > 0).to(torch.long)
        sample = [x, y]
        return sample


class DummyClassificationDataLoader(RideClassificationDataset):
    @staticmethod
    def configs() -> Configs:
        c = Configs.collect(DummyClassificationDataLoader)
        c.add(
            name="input_shape",
            type=int,
            default=10,
            strategy="constant",
            description="Input shape for data.",
        )
        return c

    def validate_attributes(self):
        RideClassificationDataset.validate_attributes(self)
        for attr in DummyRegressionDataLoader.configs().names:
            assert some(
                self, f"hparams.{attr}"
            ), f"ClassificationLifecycle should define `hparams.{attr}` but none was found."

    def __init__(self, hparams):
        num_workers = 1
        self.input_shape = (hparams.input_shape,)
        self.output_shape = 2
        self.classes = ["0", "1"]

        self._train_dataloader = DataLoader(
            OverOneDataset(self.input_shape, num_samples=65),
            batch_size=hparams.batch_size,
            num_workers=num_workers,
            shuffle=True,
            pin_memory=num_workers > 1,
        )
        self._val_dataloader = DataLoader(
            OverOneDataset(self.input_shape, num_samples=4),
            batch_size=hparams.batch_size,
            num_workers=num_workers,
            shuffle=False,
            pin_memory=num_workers > 1,
        )
        self._test_dataloader = DataLoader(
            OverOneDataset(self.input_shape, num_samples=10),
            batch_size=hparams.batch_size,
            num_workers=num_workers,
            shuffle=False,
            pin_memory=num_workers > 1,
        )

    def train_dataloader(self) -> DataLoader:
        return self._train_dataloader

    def val_dataloader(self) -> DataLoader:
        return self._val_dataloader

    def test_dataloader(self) -> DataLoader:
        return self._test_dataloader



import logging
from argparse import ArgumentParser

import pytorch_lightning as pl
import torch

from ride.core import AttributeDict, RideMixin, RideModule
from ride.optimizers import SgdOptimizer
from ride.utils.utils import DictLike

from .dummy_dataset import DummyRegressionDataLoader

pl.seed_everything(42)


class Mixin1(RideMixin):
    def __init__(self, hparams: AttributeDict):
        self.hparams.msg.append("Mixin1.__init__")

    def on_init_end(self, hparams, *args, **kwargs):
        self.hparams.msg.append("Mixin1.on_init_end")


class Mixin2(RideMixin):
    def __init__(self, hparams: AttributeDict):
        self.hparams.msg.append("Mixin2.__init__")

    def on_init_end(self, hparams, *args, **kwargs):
        self.hparams.msg.append("Mixin2.on_init_end")


class InitOrderModule(
    RideModule,
    Mixin1,
    Mixin2,
):
    def __init__(self, hparams: DictLike):
        self.hparams.msg.append("InitOrderModule.__init__")
        self.input_shape = (1,)
        self.output_shape = (1,)


def test_init_order():

    module = InitOrderModule({"msg": []})

    assert module.hparams.msg == [
        "Mixin1.__init__",
        "Mixin2.__init__",
        "InitOrderModule.__init__",
        "Mixin1.on_init_end",
        "Mixin2.on_init_end",
    ]


class DummyModule(
    RideModule,
    SgdOptimizer,
    DummyRegressionDataLoader,
):
    # Not needed:
    # @staticmethod
    # def configs() -> Configs:
    #     return Configs.collect(DummyModule)

    def __init__(self):
        self.lin = torch.nn.Linear(
            self.input_shape[0],  # from DummyRegressionDataLoader
            self.output_shape,  # from DummyRegressionDataLoader
        )
        # Alternative way of specifying loss:
        self.loss = torch.nn.functional.mse_loss

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.lin(x)


def test_init_only_self():
    """Test if modules initialises if """
    DummyModule()


def test_nested_init_hparams():
    """Test if modules initialises when hparams is also given"""

    class DummyModuleNested(DummyModule):
        def __init__(self, hparams, something_else=42):
            # super().__init__(hparams)
            self.something_else = something_else

    parser = DummyModuleNested.configs().add_argparse_args(ArgumentParser())
    args, _ = parser.parse_known_args()
    args = apply_standard_args(args)

    module = DummyModuleNested(args)
    assert module.something_else == 42

    module = DummyModuleNested(args, 1337)
    assert module.something_else == 1337

    module = DummyModuleNested(args, something_else=13)
    assert module.something_else == 13


def test_training_step_added_automatically():
    """Test if modules adds lifecycle automatically"""

    parser = DummyModule.configs().add_argparse_args(ArgumentParser())
    args, _ = parser.parse_known_args()
    args = apply_standard_args(args)
    module = DummyModule(args)
    assert module.loss.__name__ == "mse_loss"
    batch_size = 2
    x = torch.ones((batch_size, *module.input_shape), dtype=torch.float)
    y = torch.ones((batch_size, module.output_shape), dtype=torch.float)
    step_output = module.training_step((x, y), batch_idx=0)

    assert type(step_output["loss"]) is torch.Tensor
    assert step_output["pred"].shape == y.shape
    assert torch.equal(step_output["target"], y)


def apply_standard_args(args):
    args.max_epochs = 3
    args.gpus = 0
    args.checkpoint_callback = True
    args.performance_metric = "loss"
    args.id = "automated_test"
    args.test_ensemble = 0
    args.loss = "mse_loss"
    return args


def test_module_with_no_forward_warns(caplog):
    """Test if modules initialises when hparams is also given"""
    caplog.clear()
    with caplog.at_level(logging.WARNING):

        class DummyModuleNoForward(
            RideModule,
            SgdOptimizer,
            DummyRegressionDataLoader,
        ):
            def __init__(self):
                self.lin = torch.nn.Linear(
                    self.input_shape[0],  # from DummyRegressionDataLoader
                    self.output_shape,  # from DummyRegressionDataLoader
                )

            # Missing on purpose:
            # def forward(self, x: torch.Tensor) -> torch.Tensor:
            #     return self.lin(x)

    assert len(caplog.messages) == 1
    assert "forward" in caplog.text



""" main.py
    Main entry-point for the Ride main wrapper.
    For logging to be formatted consistently, this file should be imported prior to other libraries

   isort:skip_file
"""

# Monkey-patch logger to enforce consistent look across libraries
import logging

original_getLogger = logging.getLogger


def patched_getLogger(name: str = None):
    if name:
        name = name.split(".")[0]  # Get chars before '.'
    if name == "pytorch_lightning":
        name = "lightning"
    return original_getLogger(name)


logging.getLogger = patched_getLogger

from ride.utils.logging import getLogger, init_logging, style_logging  # noqa: E402

style_logging()


from argparse import ArgumentParser  # noqa: E402
from pathlib import Path  # noqa: E402
from typing import Any, Callable, Type, List  # noqa: E402
import yaml  # noqa: E402
import platform  # noqa: E402
from pytorch_lightning import Trainer, seed_everything  # noqa: E402
from ride.core import Configs, RideModule  # noqa: E402
from ride.hparamsearch import Hparamsearch  # noqa: E402
from ride.logging import experiment_logger  # noqa: E402
from ride.runner import Runner  # noqa: E402
from ride.utils.checkpoints import find_checkpoint  # noqa: E402
from ride.utils.env import LOGS_PATH  # noqa: E402
from ride.utils.io import bump_version, dump_yaml  # noqa: E402
from ride.utils.utils import attributedict  # noqa: E402
from pytorch_lightning.utilities.parsing import AttributeDict  # noqa: E402

logger = getLogger(__name__)


class Main:
    """Complete main programme for the lifecycle of a machine learning project

    Usage:
        Main(YourRideModule).argparse()
    """

    def __init__(self, Module: Type[RideModule]):
        self.Module = Module
        self.module_name = Module.__name__
        self.runner = Runner(self.Module)
        self.hparamsearch = Hparamsearch(self.Module)

    def argparse(
        self,
        args: List[str] = None,
        run=True,
    ):
        parser = ArgumentParser(add_help=True)

        # Top level commands
        top_level_commands_and_descriptions = [
            (
                "hparamsearch",
                "Run hyperparameter search. The best hyperparameters will be used for subsequent lifecycle methods",
            ),
            ("train", "Run model training"),
            ("validate", "Run model evaluation on validation set"),
            ("test", "Run model evaluation on test set"),
            ("profile_model", "Profile the model"),
        ]
        prog_flow_parser = parser.add_argument_group(
            "Flow",
            description="Commands that control the top-level flow of the programme.",
        )
        for c, d in top_level_commands_and_descriptions:
            prog_flow_parser.add_argument(f"--{c}", action="store_true", help=d)

        # General settings
        gen_settings_parser = parser.add_argument_group(
            "General",
            description="Settings that apply to the programme in general.",
        )
        gen_configs = Configs()
        gen_configs.add(
            "id",
            type=str,
            default="unnamed",
            description="Identifier for the run. If not specified, the current timestamp will be used",
        )
        gen_configs.add(
            "seed",
            type=int,
            default=123,
            description="Global random seed",
        )
        gen_configs.add(
            "logging_backend",
            type=str,
            choices=("tensorboard", "wandb"),
            default="tensorboard",
            description="Type of experiment logger",
        )
        gen_configs.add(
            "from_hparams_file",
            type=str,
            default=None,
            description="Path to JSON hparams file",
        )
        gen_configs.add(
            name="optimization_metric",
            default="loss",
            type=str,
            choices=self.Module.metric_names(),
            description="Name of the performance metric that should be optimized",
        )
        gen_configs.add(
            "monitor_lr",
            type=int,
            default=1,
            choices=[0, 1],
            description="Whether to monitor and log learning rate",
        )
        gen_configs.add(
            "checkpoint_every_n_steps",
            type=int,
            default=0,
            description="Save models checkpoint every N steps independent of epoch and validation cycle. If `0`, this feature is unused",
        )
        gen_configs.add(
            "profile_model_num_runs",
            type=int,
            default=0,
            description="Number of runs to perform when profiling model. If `0`, model will be profiled to max 10 seconds.",
        )
        gen_settings_parser = gen_configs.add_argparse_args(gen_settings_parser)

        # Pytorch Lightning args
        pl_parser = parser.add_argument_group(
            "Pytorch Lightning",
            description="Settings inherited from the pytorch_lightning.Trainer",
        )
        pl_configs = Configs.from_argument_parser(
            Trainer.add_argparse_args(ArgumentParser(add_help=False))
        )
        pl_parser = pl_configs.add_argparse_args(pl_parser)

        # Hparamsearch
        hparamsearch_parser = parser.add_argument_group(
            "Hparamsearch",
            description="Settings associated with hyperparameter optimisation",
        )
        hparamsearch_parser = (
            self.hparamsearch.configs() - gen_configs
        ).add_argparse_args(hparamsearch_parser)

        # Module args
        module_parser = parser.add_argument_group(
            "Module",
            description="Settings associated with the Module",
        )
        module_parser = (
            self.Module.configs() - gen_configs - pl_configs
        ).add_argparse_args(module_parser)

        if run:
            parsed_args = parser.parse_args(args)
            return self.main(parsed_args)
        else:
            return parser

    def main(self, args: AttributeDict):  # noqa: C901
        args = attributedict(args)

        seed_everything(args.seed)
        self.log_dir = experiment_logger(
            args.id, args.logging_backend, self.Module
        ).log_dir
        init_logging(
            self.log_dir,
            args.logging_backend,
        )

        if getattr(args, "num_workers", False) and platform.system() == "Windows":
            logger.warning(
                f"You have requested num_workers={args.num_workers} on Windows, but currently 0 is recommended (see https://stackoverflow.com/a/59680818)"
            )

        results = []

        if not args.default_root_dir:
            args.default_root_dir = str(LOGS_PATH)

        save_results = make_save_results(self.log_dir)

        if args.resume_from_checkpoint:
            args.resume_from_checkpoint = find_checkpoint(args.resume_from_checkpoint)

        if args.from_hparams_file:
            logger.info(f"Loading hyperparameters from {args.from_hparams_file}")
            args = Hparamsearch.load(
                args.from_hparams_file,
                old_args=args,
                Cls=self.Module,
                auto_scale_lr=True,
            )

        if args.hparamsearch:
            hprint("Searching for optimal model hyperparameters")
            best_hparams = self.hparamsearch.run(args)
            if not best_hparams:
                raise RuntimeError("Hparamsearch was unable to identy best hparams")
            logger.info("Search found these hyperparameters:")
            dprint(best_hparams)
            results.append(best_hparams)
            save_results(
                f"hparamsearch/{args.id}.yaml",
                best_hparams,
            )
            best_hparams_path = Hparamsearch.dump(best_hparams, identifier=args.id)
            logger.info(f"Loading best hparams from '{best_hparams_path}'")
            args = Hparamsearch.load(best_hparams_path, old_args=args)

        if args.train:
            hprint("Running training")
            self.runner.train(args)

        if args.validate:
            hprint("Running evaluation on validation set")
            val_results = self.runner.validate(args)
            dprint(val_results)
            results.append(val_results)
            save_results(
                "evaluation/validation_results.yaml",
                val_results,
            )

        if args.test:
            hprint("Running evaluation on test set")
            test_results = self.runner.test(args)
            dprint(test_results)
            results.append(test_results)
            save_results(
                "evaluation/test_results.yaml",
                test_results,
            )

        if args.profile_model:
            hprint("Profiling model")
            info = self.runner.profile_model(args, num_runs=args.profile_model_num_runs)
            dprint(info)
            results.append(info)
            save_results(
                "profiles/model_profile.yaml",
                info,
            )
            save_results(
                "profiles/model_hparams.yaml",
                vars(args),
            )

        return results


def hprint(msg: str):
    """Message header print

    Args:
        msg (str): Message to be printed
    """
    logger.info(f"🚀 {msg}")


def dprint(d: dict):
    logger.info(yaml.dump({"Results": d}))


def make_save_results(root_path: str) -> Callable[[str, Any], None]:
    root_path = Path(root_path)

    def bump_version_and_save(relative_path: str, data):
        nonlocal root_path
        path = bump_version(root_path / relative_path)
        logger.info("Saving " + str(path))
        dump_yaml(path, data)

    return bump_version_and_save



import os
from pathlib import Path

from dotenv import find_dotenv, load_dotenv

load_dotenv(find_dotenv())

ROOT_PATH = Path(os.getenv("ROOT_PATH", default=os.getcwd()))
DATASETS_PATH = ROOT_PATH / os.getenv("DATASETS_PATH", default="datasets")
LOGS_PATH = ROOT_PATH / os.getenv("LOGS_PATH", default="logs")
RUN_LOGS_PATH = LOGS_PATH / "run_logs"
TUNE_LOGS_PATH = LOGS_PATH / "tune_logs"
CACHE_PATH = ROOT_PATH / os.getenv("CACHE_PATH", default=".cache")
LOG_LEVEL = os.getenv("LOG_LEVEL", default="INFO")

DATASETS_PATH.mkdir(exist_ok=True)
LOGS_PATH.mkdir(exist_ok=True)
RUN_LOGS_PATH.mkdir(exist_ok=True)
TUNE_LOGS_PATH.mkdir(exist_ok=True)
CACHE_PATH.mkdir(exist_ok=True)

NUM_CPU = os.cpu_count() or 1



import platform
import shutil
from argparse import ArgumentParser
from pathlib import Path
from typing import Tuple

import pytest
import pytorch_lightning as pl
import torch

from ride.core import Configs, RideModule
from ride.logging import experiment_logger
from ride.optimizers import SgdOptimizer
from ride.runner import Runner
from ride.utils.utils import AttributeDict

from .dummy_dataset import DummyRegressionDataLoader

pl.seed_everything(42)


class DummyModule(RideModule, DummyRegressionDataLoader, SgdOptimizer):
    def __init__(self, hparams):
        self.l1 = torch.nn.Linear(self.input_shape[0], self.hparams.hidden_dim)
        self.l2 = torch.nn.Linear(self.hparams.hidden_dim, self.output_shape)
        self.loss = torch.nn.functional.mse_loss

    def forward(self, x):
        x = x.view(x.size(0), -1)
        x = torch.relu(self.l1(x))
        x = torch.relu(self.l2(x))
        return x

    @staticmethod
    def configs():
        c = Configs()
        c.add(
            name="hidden_dim",
            type=int,
            default=128,
            strategy="choice",
            choices=[128, 256, 512, 1024],
            description="Number of hiden units.",
        )
        return c


@pytest.fixture()  # scope="module"
def runner_and_args() -> Tuple[Runner, AttributeDict]:
    r = Runner(DummyModule)
    parser = r.Module.configs().add_argparse_args(ArgumentParser())
    args, _ = parser.parse_known_args()
    args.max_epochs = 1
    args.gpus = 0
    args.checkpoint_callback = True
    args.optimization_metric = "loss"
    args.id = "automated_test"
    args.test_ensemble = 0
    args.checkpoint_every_n_steps = 0
    args.monitor_lr = 0
    args.auto_lr_find = 0
    args.auto_scale_batch_size = 0
    args.num_workers = 1
    args.batch_size = 4
    args.logging_backend = "tensorboard"

    return r, args


class TestRunner:
    def test_init(self):
        class NotARideModule:
            meaningoflife = 42

        with pytest.raises(AssertionError):
            r = Runner(NotARideModule)  # type:ignore

        r = Runner(DummyModule)
        assert r.Module is not None

    def test_train(self, runner_and_args: Tuple[Runner, AttributeDict]):
        """Test that training works"""
        runner, args = runner_and_args
        runner.train(args)

    def test_train_auto_scale_batch_size(
        self, runner_and_args: Tuple[Runner, AttributeDict]
    ):
        runner, args = runner_and_args
        args.auto_scale_batch_size = "power"
        runner.train(args)

    def test_train_auto_lr_find(self, runner_and_args: Tuple[Runner, AttributeDict]):
        runner, args = runner_and_args
        args.auto_lr_find = True
        runner.train(args)

    def test_validate(self, runner_and_args: Tuple[Runner, AttributeDict]):
        """Test that evaluation using validation set works
        Implicitely tests Runner.evaluate
        """
        runner, args = runner_and_args
        result = runner.validate(args)
        assert result["val/loss"]

    def test_test(self, runner_and_args: Tuple[Runner, AttributeDict]):
        """Test that evaluation using test set works
        Implicitely tests Runner.evaluate
        """
        runner, args = runner_and_args
        result = runner.test(args)
        assert result["test/loss"]

    def test_train_and_val(self, runner_and_args: Tuple[Runner, AttributeDict]):
        """Test that training and evaluation using validation set works"""
        runner, args = runner_and_args
        runner.train_and_val(args)

    def test_train_and_val_static(self, runner_and_args: Tuple[Runner, AttributeDict]):
        """Test that static version of training and evaluation using validation set works"""
        _, args = runner_and_args
        Runner.static_train_and_val(DummyModule, args)

    def test_profile_model(self, runner_and_args: Tuple[Runner, AttributeDict]):
        """Test that profiling model work"""
        runner, args = runner_and_args
        runner.profile_model(args, max_wait_seconds=1)

    def test_teardown(self, runner_and_args: Tuple[Runner, AttributeDict]):
        runner, args = runner_and_args
        d = Path(experiment_logger(args.id).log_dir).parent

        try:
            # May throw error on Windows:
            # PermissionError: [WinError 32] The process cannot access the file because it is being used by another process
            shutil.rmtree(d)
        except Exception:
            pass

        if platform.system() != "Windows":
            assert not d.exists()



import logging
import os
import socket
import subprocess
import sys
from functools import wraps
from pathlib import Path

import click
import coloredlogs
import pytorch_lightning as pl

from ride.utils.env import LOG_LEVEL
from ride.utils.utils import once

LOG_LEVELS = {"DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"}


def _process_rank():
    if pl.utilities._HOROVOD_AVAILABLE:
        import horovod.torch as hvd

        hvd.init()
        return hvd.rank()
    else:
        return pl.utilities.rank_zero_only.rank


process_rank = _process_rank()


def if_rank_zero(fn):
    @wraps(fn)
    def wrapped(*args, **kwargs):
        global process_rank
        if process_rank == 0:
            fn(*args, **kwargs)

    return wrapped


def getLogger(name, log_once=False):
    name = name.split(".")[0]  # Get chars before '.'
    if name not in {"wandb", "lightning", "ride", "datasets", "models"}:
        name = click.style(name, fg="white", bold=True)
    logger = logging.getLogger(name)
    if log_once:
        logger._log = once(logger._log)
    logger._log = if_rank_zero(logger._log)
    return logger


logger = getLogger(__name__)


_ansi_colors = {
    "black": 30,
    "red": 31,
    "green": 32,
    "yellow": 33,
    "blue": 34,
    "magenta": 35,
    "cyan": 36,
    "white": 37,
    "reset": 39,
    "bright_black": 90,
    "bright_red": 91,
    "bright_green": 92,
    "bright_yellow": 93,
    "bright_blue": 94,
    "bright_magenta": 95,
    "bright_cyan": 96,
    "bright_white": 97,
}
_ansi_reset_all = "\033[0m"


def style(  # noqa: C901
    text,
    fg=None,
    bg=None,
    bold=None,
    dim=None,
    underline=None,
    blink=None,
    reverse=None,
    reset=True,
):
    """Styles a text with ANSI styles and returns the new string.  By
    default the styling is self contained which means that at the end
    of the string a reset code is issued.  This can be prevented by
    passing ``reset=False``.

    This is a modified version of the one found in `click` https://click.palletsprojects.com/en/7.x/

    Examples::

        logger.info(style('Hello World!', fg='green'))
        logger.info(style('ATTENTION!', blink=True))
        logger.info(style('Some things', reverse=True, fg='cyan'))

    Supported color names:

    * ``black`` (might be a gray)
    * ``red``
    * ``green``
    * ``yellow`` (might be an orange)
    * ``blue``
    * ``magenta``
    * ``cyan``
    * ``white`` (might be light gray)
    * ``bright_black``
    * ``bright_red``
    * ``bright_green``
    * ``bright_yellow``
    * ``bright_blue``
    * ``bright_magenta``
    * ``bright_cyan``
    * ``bright_white``
    * ``reset`` (reset the color code only)

    :param text: the string to style with ansi codes.
    :param fg: if provided this will become the foreground color.
    :param bg: if provided this will become the background color.
    :param bold: if provided this will enable or disable bold mode.
    :param dim: if provided this will enable or disable dim mode.  This is
                badly supported.
    :param underline: if provided this will enable or disable underline.
    :param blink: if provided this will enable or disable blinking.
    :param reverse: if provided this will enable or disable inverse
                    rendering (foreground becomes background and the
                    other way round).
    :param reset: by default a reset-all code is added at the end of the
                  string which means that styles do not carry over.  This
                  can be disabled to compose styles.
    """
    bits = []
    if fg:
        try:
            bits.append("\033[{}m".format(_ansi_colors[fg]))
        except KeyError:
            raise TypeError("Unknown color '{}'".format(fg))
    if bg:
        try:
            bits.append("\033[{}m".format(_ansi_colors[bg] + 10))
        except KeyError:
            raise TypeError("Unknown color '{}'".format(bg))
    if bold is not None:
        bits.append("\033[{}m".format(1 if bold else 22))
    if dim is not None:
        bits.append("\033[{}m".format(2 if dim else 22))
    if underline is not None:
        bits.append("\033[{}m".format(4 if underline else 24))
    if blink is not None:
        bits.append("\033[{}m".format(5 if blink else 25))
    if reverse is not None:
        bits.append("\033[{}m".format(7 if reverse else 27))
    bits.append(text)
    if reset:
        bits.append(_ansi_reset_all)
    return "".join(bits)


def style_logging():
    assert LOG_LEVEL in LOG_LEVELS, f"Specified LOG_LEVEL should be one of {LOG_LEVELS}"

    # Installing coloredlogs during tests doesn't work
    if "pytest" not in sys.modules:
        coloredlogs.install(
            level=LOG_LEVEL,
            fmt="%(name)s: %(message)s",
            level_styles={
                "debug": {"color": "white", "faint": True},
                "warning": {"bold": True},
                "error": {"color": "red", "bold": True},
            },
        )

    # Block pytorch_lightning from writing directly to stdout
    lightning_logger = getattr(pl, "_logger", logging.getLogger("lightning"))
    lightning_logger.handlers = []
    lightning_logger.propagate = bool(process_rank == 0)

    # Set coloring
    lightning_logger.name = style(lightning_logger.name, fg="yellow", bold=True)

    ride_logger = logging.getLogger("ride")
    ride_logger.name = style(ride_logger.name, fg="cyan", bold=True)

    datasets_logger = logging.getLogger("datasets")
    datasets_logger.name = style(datasets_logger.name, fg="magenta", bold=True)

    models_logger = logging.getLogger("models")
    models_logger.name = style(models_logger.name, fg="green", bold=True)

    # Block matplotlib debug logger
    matplotlib_logger = logging.getLogger("matplotlib")
    matplotlib_logger.setLevel(
        logging.INFO if LOG_LEVEL == "DEBUG" else getattr(logging, LOG_LEVEL)
    )


def init_logging(logdir: str = None, logging_backend: str = "tensorboard"):
    if not logdir:
        return

    # Add root handler for redirecting run output to file
    os.makedirs(logdir, exist_ok=True)
    getLogger("").addHandler(logging.FileHandler(Path(logdir) / "run.log"))

    # Write basic environment info to logs
    logger.info(f"Running on host {style(socket.gethostname(), fg='yellow')}")

    try:
        git_repo = (
            subprocess.check_output(
                "git config --get remote.origin.url",
                shell=True,
                stderr=subprocess.STDOUT,
            )
            .decode("utf-8")
            .strip()
        )
        git_tag = (
            subprocess.check_output(
                "git rev-parse HEAD", shell=True, stderr=subprocess.STDOUT
            )
            .decode("utf-8")
            .strip()
        )
        git_msg = style(
            f"{git_repo.replace('.git','')}/tree/{git_tag}", fg="blue", bold=False
        )
        logger.info(f"⭐️ View project repository at {git_msg}")
    except subprocess.CalledProcessError:
        pass

    logger.info(f"Run data is saved locally at {logdir}")
    logger.info(f"Logging using {style(logging_backend.capitalize(), fg='yellow')}")



import pytest  # noqa: F401

from ride import metrics  # noqa: F401


def test_BaseMetricMixin():
    """Test that base class has sensible defaults"""
    assert False


class TestTopKAccuracyMetric:
    def test_init(self):
        """Test intialisation"""
        assert False

    def test_validate_attributes(self):
        """Test attribute validation"""
        assert False

    def test_metrics_step(self):
        """Test metrics step"""
        assert False

    def test_metrics_epoch(self):
        """Test metrics step"""
        assert False


class TestFlopsMetricMixin:
    def test_init(self):
        """Test intialisation"""
        assert False


class FlopsWeightedAccuracyMetric:
    def test_init(self):
        """Test intialisation"""
        assert False

    def test_validate_attributes(self):
        """Test attribute validation"""
        assert False

    def test_metrics_step(self):
        """Test metrics step"""
        assert False

    def test_metrics_epoch(self):
        """Test metrics step"""
        assert False


def test_topk_accuracies():
    """Test topk_accuracies"""
    assert False


def test_accuracy():
    """Test accuracy"""
    assert False


def test_topk_errors():
    """Test topk_errors"""
    assert False


def test_gflops():
    """Test computation of gflops"""
    assert False


def test_activations():
    """Test computation of activations"""
    assert False


def test_params_count():
    """Test computation of the number of parameters"""
    assert False



import json
from pathlib import Path
from typing import Any

import numpy as np
import yaml
from torch import Tensor


def bump_version(path: Path):
    """Bumps the version number for a path if it already exists"""
    if not path.exists():
        return path

    # Check for already bumped versions
    try:
        prev_version = max(
            map(
                int,
                filter(
                    lambda s: s.isdigit(),
                    [f.stem.split("_")[-1] for f in path.parent.glob(f"{path.stem}*")],
                ),
            )
        )
        new_version = prev_version + 1
    except ValueError:  # max() arg is an empty sequence
        new_version = 1

    new_name = f"{path.stem}_{new_version}{path.suffix}"
    return path.parent / new_name


def load_structured_data(path: Path):
    suffix = path.suffix
    assert suffix in {".json", ".yml", ".yaml"}
    if suffix == ".yml":
        suffix = ".yaml"
    d = {".json": load_json, ".yaml": load_yaml}[suffix](path)
    return d


def dump_yaml(path: Path, data: Any):
    path.parent.mkdir(parents=True, exist_ok=True)
    with open(str(path), "w") as f:
        yaml.dump(data, f, sort_keys=True)


def load_yaml(path: Path) -> Any:
    with open(path, "r") as f:
        data = yaml.load(f, Loader=yaml.FullLoader)
    return data


def dump_json(path: Path, data: Any):
    path.parent.mkdir(parents=True, exist_ok=True)
    with open(str(path), "w") as f:
        json.dump(data, f, cls=NpJsonEncoder, sort_keys=True, indent=2)


def load_json(path: Path) -> Any:
    with open(path, "r") as f:
        data = json.load(f)
    return data


class NpJsonEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):  # type:ignore
            return int(obj)
        elif isinstance(obj, np.floating):  # type:ignore
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return super(NpJsonEncoder, self).default(obj)


def float_representer(dumper: yaml.Dumper, value: float):
    text = "{0:.9f}".format(value)
    return dumper.represent_scalar("tag:yaml.org,2002:float", text)


def tensor_representer(dumper: yaml.Dumper, data: Tensor):
    assert type(data) == Tensor
    if data.shape:
        return dumper.represent_sequence("tag:yaml.org,2002:float", data.tolist())
    else:
        return float_representer(dumper, data.item())


yaml.add_representer(float, float_representer)
yaml.add_representer(Tensor, tensor_representer)
